/* -*-c-*-
 *
 * $Id: identify.c,v 1.3 1999/07/26 23:26:21 mdw Exp $
 *
 * Identifies and logs the client of a connection
 *
 * (c) 1999 Straylight/Edgeware
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of the `fw' port forwarder.
 *
 * `fw' is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * `fw' is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with `fw'; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: identify.c,v $
 * Revision 1.3  1999/07/26 23:26:21  mdw
 * Minor modifications for new design.
 *
 * Revision 1.2  1999/07/03 13:56:59  mdw
 * Log connections to syslog or stderr as appropriate.
 *
 * Revision 1.1.1.1  1999/07/01 08:56:23  mdw
 * Initial revision.
 *
 */

/*----- Header files ------------------------------------------------------*/

#include "config.h"

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <syslog.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include <mLib/alloc.h>
#include <mLib/conn.h>
#include <mLib/dstr.h>
#include <mLib/report.h>
#include <mLib/sel.h>
#include <mLib/selbuf.h>
#include <mLib/str.h>

#include "bres.h"
#include "fw.h"
#include "ident.h"
#include "identify.h"

/*----- Magic numbers -----------------------------------------------------*/

#define TIMEOUT 15			/* Seconds to wait for answers */

/*----- Data structures ---------------------------------------------------*/

/* --- Structure to track the progress of an identification --- */

typedef struct id {
  id_req q;				/* Copy of client's request block */
  time_t when;				/* When the connection occurred */
  conn c;				/* Connection selector */
  unsigned state;			/* Current state of the world */
  bres_client r;			/* Backgd resolver client block */
  char host[64];			/* Resolved hostname */
  char user[32];			/* Authenticated client user */
  sel_timer t;				/* Timeout selector */
  selbuf id;				/* Reader for the RFC931 client */
} id;

#define S_HOST 1u			/* Read the hostname from resolver */
#define S_USER 2u			/* Read the username from RFC931 */
#define S_UCONN 4u			/* Connected to remote RFC931 */
#define S_TIMER 8u			/* Timeout has completed */

/*----- Main code ---------------------------------------------------------*/

/* --- @id_done@ --- *
 *
 * Arguments:	@id *i@ = pointer to identification block
 *
 * Returns:	---
 *
 * Use:		Finishes with an identification block.
 */

static void id_done(id *i)
{
  /* --- Close down the various dependent bits --- */

  if (!(i->state & S_HOST))
    bres_abort(&i->r);
  if (!(i->state & S_UCONN))
    conn_kill(&i->c);
  else if (!(i->state & S_USER))
    selbuf_disable(&i->id);
  if (!(i->state & S_TIMER))
    sel_rmtimer(&i->t);

  /* --- Report the final result --- */

  fw_log(i->when, "[%s] %s from %s@%s [%s]",
	 i->q.desc, i->q.act,
	 i->user, i->host, inet_ntoa(i->q.rsin.sin_addr));

  /* --- Dispose of the block --- */

  REFFD_DEC(i->q.r);
  free(i);
}

/* --- @id_res@ --- *
 *
 * Arguments:	@const char *host@ = name of the resolved host
 *		@void *vp@ = pointer to identification block
 *
 * Returns:	---
 *
 * Use:		Responds to a completed reverse name resolution.
 */

static void id_res(const char *host, void *vp)
{
  id *i = vp;
  str_sanitize(i->host, host, sizeof(i->host));
  i->state |= S_HOST;
  if (i->state & S_USER)
    id_done(i);
}

/* --- @id_ident@ --- *
 *
 * Arguments:	@char *p@ = pointer to string read from server
 *		@void *vp@ = pointer to identification block
 *
 * Returns:	---
 *
 * Use:		Responds to a line read from the remote RFC931 server.
 */

static void id_ident(char *p, void *vp)
{
  id *i = vp;

  /* --- Get rid of the connection --- */

  i->state |= S_USER;
  selbuf_disable(&i->id);
  close(i->id.reader.fd);

  /* --- Read the information from the returned line --- */

  if (p) {
    ident idbuf;
    ident_parse(p, &idbuf);
    if (idbuf.type == ident_userid)
      str_sanitize(i->user, idbuf.u.userid.user, sizeof(i->user));
  }

  /* --- Maybe finish off this identification --- */

  if (i->state & S_HOST)
    id_done(i);
}

/* --- @id_conn@ --- *
 *
 * Arguments:	@int fd@ = file descriptor connected
 *		@void *vp@ = pointer to identification block
 *
 * Returns:	---
 *
 * Use:		Responds to a completed connection to the remote RFC931
 *		server.
 */

static void id_conn(int fd, void *vp)
{
  id *i = vp;

  if (fd == -1) {
    i->state |= S_USER | S_UCONN;
    if (i->state & S_HOST)
      id_done(i);
  } else {
    dstr d = DSTR_INIT;
    dstr_putf(&d, "%u, %u\n",
	      ntohs(i->q.rsin.sin_port), ntohs(i->q.lsin.sin_port));
    write(fd, d.buf, d.len);
    dstr_destroy(&d);
    i->state |= S_UCONN;
    selbuf_init(&i->id, sel, fd, id_ident, i);
  }
}

/* --- @id_timer@ --- *
 *
 * Arguments:	@struct timeval *tv@ = pointer to the current time
 *		@void *vp@ = pointer to identification block
 *
 * Returns:	---
 *
 * Use:		Times an identification job out.
 */

static void id_timer(struct timeval *tv, void *vp)
{
  id *i = vp;
  id_done(i);
}

/* --- @identify@ --- *
 *
 * Arguments:	@const id_req *q@ = pointer to request block
 *
 * Returns:	---
 *
 * Use:		Starts a background ident lookup and reverse-resolve job
 *		which will, eventually, report a message to the system log.
 */

void identify(const id_req *q)
{
  id *i;

  /* --- Initialize the block with stuff --- */

  i = xmalloc(sizeof(*i));
  i->q = *q;
  REFFD_INC(i->q.r);

  str_sanitize(i->host, inet_ntoa(q->rsin.sin_addr), sizeof(i->host));
  strcpy(i->user, "<ANONYMOUS>");
  i->state = 0;
  i->when = time(0);

  /* --- Set up the connection to the identity server --- */

  {
    int fd;
    struct sockaddr_in sin;

    if ((fd = socket(PF_INET, SOCK_STREAM, 0)) < 0)
      id_conn(-1, i);
    else {
      sin.sin_family = AF_INET;
      sin.sin_addr = q->lsin.sin_addr;
      sin.sin_port = 0;
      if (bind(fd, (struct sockaddr *)&sin, sizeof(sin))) {
	close(fd);
	id_conn(-1, i);
      } else {
	int opt = 1;
	sin.sin_family = AF_INET;
	sin.sin_addr = q->rsin.sin_addr;
	sin.sin_port = htons(113);
	setsockopt(fd, SOL_SOCKET, SO_OOBINLINE, &opt, sizeof(opt));
	conn_init(&i->c, sel, fd,
		  (struct sockaddr *)&sin, sizeof(sin),
		  id_conn, i);
      }
    }
  }

  /* --- Set up the name resolver --- */

  bres_resolve(&i->r, q->rsin.sin_addr, id_res, i);

  /* --- Set up the time limiter --- */

  {
    struct timeval tv;
    gettimeofday(&tv, 0);
    tv.tv_sec += TIMEOUT;
    sel_addtimer(sel, &i->t, &tv, id_timer, i);
  }
}

/*----- That's all, folks -------------------------------------------------*/
